structure Exceptions :> NUBOX =
struct

    structure Support = SupportSet

    structure Types = 
	ObjectTypes (structure Support = Support)

    structure Absyn = 
	AbstractSyntax (structure Support = SupportSet
			structure Types = Types)
	
    structure Parse =
	Parsing (structure Absyn = Absyn)
	
    structure Ctx = 
	Context (structure Absyn = Absyn)
	
    structure Typecheck = 
	Typechecking (structure Absyn = Absyn
		      structure Ctx = Ctx
		      val lang = Language.EXCPTN)
	
	
    exception RuntimeError
    
    type var = Var.var
    open Absyn


    datatype cterm = 
	CEmpty 
      | CNat of int
      | CTrue 
      | CFalse
      | CNil
      | CCons of cterm * cterm 
      | CFn of env * var * term
      | CFix of env * var * var * term
      | CPair of cterm * cterm
      | CBox of env * term
      | CNu of env * var * term

    and value = 
	Val of cterm 
      | Exp of env * term 

	
    withtype env = value Var.VarMap.map


    exception Internal of var * cterm


    fun valToTerm (v) =
	(case v of
	     Val ct => toTerm ct
	   | Exp (env, t) => envToTerm (env, t))

    and valToString (v) = termToString (valToTerm v)

    and envToSubst (e) = Var.VarMap.map valToTerm e
	
    and envToTerm (env, t) = 
	let val s = envToSubst env
	in
	    substTerm (s, t)
	end

    and toTerm (ct : cterm) : term =
	(case ct of
	     CEmpty => Empty
	   | CNat n => Nat n
	   | CTrue => True
	   | CFalse => False
	   | CNil => Nil
	   | CCons (ct1, ct2) => 
		 Cons (toTerm ct1, toTerm ct2)
	   | CFn (boundVars, x, e) => 
		 let val env = (#1 (Var.VarMap.remove (boundVars, x))
				handle _ => boundVars)
		     val e' = envToTerm (env, e)
		 in
		     Fn (x, e')
		 end
	   | CFix (boundVars, f, x, e) => 
		 let val env' = (#1 (Var.VarMap.remove (boundVars, f))
				 handle _ => boundVars)
		     val env = (#1 (Var.VarMap.remove (env', x))
				handle _ => env')
		     val e' = envToTerm (env, e)
		 in
		     Fix (f, x, e')
		 end
	   | CPair (ct1, ct2) => 
		 Pair (toTerm ct1, toTerm ct2)
	   | CBox (boundVars, e) => 
		 let val e' = envToTerm (boundVars, e)
		 in
		     Box (e')
		 end
	   | CNu (boundVars, x, e) => 
		 let val env = (#1 (Var.VarMap.remove (boundVars, x))
				 handle _ => boundVars)
		     val e' = envToTerm (env, e)
		 in
		     Nu (x, e)
		 end)



    (***
     each evaluation function takes a static environment 
     providing values and/or expressions for variables 
     ***)
     

    fun evalTerm (varEnv : env, e : term) : cterm =
	(case e of
	     Empty => CEmpty
	   | Print e' => 
		 let val v' = evalTerm (varEnv, e')
		     val v'' = toTerm v'
		     val _ = print ((termToString v'')^"\n")
		 in
		     CEmpty
		 end
	   | Nat x => CNat x
	   | True => CTrue
	   | False => CFalse
	   | Nil => CNil
	   | Cons (e1, e2) => 
		 let val v1 = evalTerm (varEnv, e1)
		     val v2 = evalTerm (varEnv, e2)
		 in
		     CCons (v1, v2)
		 end
	   | Car (e1) => 
		 let val CCons (hd, tl) = evalTerm (varEnv, e1)
		 in
		     hd
		 end

	   | Cdr (e1) => 
		 let val CCons (hd, tl) = evalTerm (varEnv, e1)
		 in
		     tl
		 end

	   | Var x => 
		 (case Var.VarMap.find (varEnv, x) of
		      SOME (Val c) => c
		    | SOME (Exp (env', e')) => evalTerm (env', e')
		    | NONE => raise RuntimeError)

	   | BinOp (p, e1, e2) => 
		 let val v1 = evalTerm (varEnv, e1)
		     val v2 = evalTerm (varEnv, e2)
		 in
	             case (p, v1, v2) of
			 (MINUS, CNat n1, CNat n2) => CNat (n1 - n2)
		       | (PLUS, CNat n1, CNat n2) => CNat (n1 + n2)
		       | (MULT, CNat n1, CNat n2) => CNat (n1 * n2)
		       | (EQL, e1, e2) => 
			     if termEq (toTerm e1, toTerm e2) then CTrue else CFalse
		 end

	   | If (c, e1, e2) => 
		 let val c' = evalTerm (varEnv, c)
		 in
		     case c' of
			 CTrue => evalTerm (varEnv, e1)
		       | CFalse => evalTerm (varEnv, e2)
		 end
		 
	   | Fn (x, e') => CFn (varEnv, x, e')

	   | Fix (f, x, e') => CFix (varEnv, f, x, e')
      
	   | App (e1, e2) => 
		 let val v1 = evalTerm (varEnv, e1)
		 in 
		     case v1 of
			 CFn (env', x, e3) => 
			     let val v2 = evalTerm (varEnv, e2)
				 val env'' = Var.VarMap.insert (env', x, Val v2)
			     in
				 evalTerm (env'', e3)
			     end
		       | CFix (env', f, x, e3) => 
			     let val v2 = evalTerm (varEnv, e2)
				 val env2 = Var.VarMap.insert (env', f, Val v1)
				 val env3 = Var.VarMap.insert (env2, x, Val v2)
			     in
				 evalTerm (env3, e3)
			     end
		       | _ => raise RuntimeError
		 end

	   | Pair (e1, e2) => 
		 let val v1 = evalTerm (varEnv, e1)
		     val v2 = evalTerm (varEnv, e2)
		 in
		     CPair (v1, v2)
		 end
		 
	   | Fst e' => 
		 let val v' = evalTerm (varEnv, e')
		 in
		     case v' of
			 CPair (v1, v2) => v1
		       | _ => raise RuntimeError
		 end

	   | Snd e' => 
		 let val v' = evalTerm (varEnv, e')
		 in
		     case v' of
			 CPair (v1, v2) => v2
		       | _ => raise RuntimeError
		 end

	   | Box e' => CBox (varEnv, e')

	   | Let (decls, e') => 
		 evalLets (varEnv, decls, e')


	   | Choose e' => 
		 let val v' = evalTerm (varEnv, e')
		 in 
		     case v' of
			 CNu (env', x, e'') => 
			     evalTerm (env', e'')
		       | _ => raise RuntimeError
		 end
	     
	   | Nu (x, e') => CNu (varEnv, x, e')

	   | Raise (x, e') => 
		 let val v' = evalTerm (varEnv, e')
		 in
		     raise Internal (x, v')
		 end

	   | Handle (handler, e') => 
		 ((evalTerm (varEnv, e'))
		 handle Internal (x, v') => 
		     evalHandler (varEnv, handler, x, v'))

	   | OfTp (e', _) => evalTerm (varEnv, e'))


    and evalHandler (varEnv, handler, x, v:cterm) : cterm =
	(case handler of
	     nil => raise Internal (x, v)
	   | ((x', y', e')::handler') => 
		 if Var.equal (x, x') then 
		     let val varEnv' = Var.VarMap.insert (varEnv, y', Val v)
		     in
			 evalTerm (varEnv', e')
		     end
		 else evalHandler (varEnv, handler', x, v))




    and evalLets (varEnv, decls, e) = 
	(case decls of
	     nil => evalTerm (varEnv, e)
	   | (NameDecl (x, tp)::ds) =>
		 let val x' = Var.derived_var x
		     val s = Var.VarMap.insert (Var.VarMap.empty, x, Var x')
		     val (s', ds') = substDecls (s, ds)
		     val e' = substTerm (s', e)
		 in
		     evalLets (varEnv, ds', e')
		 end

	   | (ValDecl (x, e')::ds) => 
		 let val v = evalTerm (varEnv, e')
		 in
		     evalLets (Var.VarMap.insert (varEnv, x, Val v), ds, e)
		 end

	   | (BoxDecl (x, e1)::ds) => 
		 let val v = evalTerm (varEnv, e1)
		 in
		     case v of
			 CBox (env', e') => 
			     evalLets (Var.VarMap.insert (varEnv, x, Exp (env', e')), ds, e)
		       | _ => raise RuntimeError
		 end)

	

    fun evalDeclsPrint (varEnv, decl) = 
	(case decl of
	     nil => print ""
	   | (NameDecl (x, tp)::ds) => 
		 let val _ = print ("name "^(Var.toString x)^"\n")
		     val x' = Var.derived_var x
		     val s = Var.VarMap.insert (Var.VarMap.empty, x, Var x')
		     val (s', ds') = substDecls (s, ds)
		 in
		     evalDeclsPrint (varEnv, ds')
		 end
	   | (ValDecl (x, e)::ds) => 
		 let val v = evalTerm (varEnv, e)
		     val _ = print ("val "^(Var.toString x)^
                                    " = "^(valToString (Val v))^"\n")
		 in
		     evalDeclsPrint (Var.VarMap.insert (varEnv, x, Val v), ds)
		 end
	   | (BoxDecl (x, e)::ds) => 
		 let val v = evalTerm (varEnv, e)
		 in
		     case v of
			 CBox (env', e') => 
			     let val _ = print ("box "^(Var.toString x)^
                                                " = box ("^(valToString (Exp (env', e')))^")\n")
			     in
				 evalDeclsPrint (Var.VarMap.insert (varEnv, x, Exp (env', e')), ds)
			     end
		       | _ => raise RuntimeError
		 end)



    fun fromFile (file : string) = 
	let val ldecls = Parse.fromFile file
	    val _ = Typecheck.typechecks ldecls
	    val _ = evalDeclsPrint (Var.VarMap.empty, map toDecl ldecls)
	in
	    ()
	end handle Error.Error msg =>  print msg 
	

end