functor AbstractSyntax(structure Support : SUPPORT
		       structure Types : TYPES
			   where type support = Support.support) :> 
    ABSTRACT_SYNTAX where Support = Support and Types = Types
=
struct
    structure Support = Support
    structure Types = Types

    type var = Var.var
    type location = Location.location
    type support = Support.support
	
    datatype binop = PLUS | MINUS | MULT | EQL

    open Types

    datatype term =
	Empty 
      | Print of term 
      |	Nat of int
      | True 
      | False 
      | Nil
      | Cons of term * term
      | Car of term 
      | Cdr of term 
      | Var of var
      | BinOp of binop * term * term
      | If of term * term * term 
      | Fn of var * term
      | Fix of var * var * term
      | App of term * term 
      | Pair of term * term 
      | Fst of term 
      | Snd of term 
      | Box of term 
      | Dia of clos
      | Let of decl list * term 
      | Choose of term
      | Nu of var * term
      | ExpSubst of (var * term) list * term
      | Raise of var * term
      | Handle of (var * var * term) list * term	
      | Shift of var * var * term
      | Reset of var * term
      | OfTp of term * tp 

    and decl = 
	NameDecl of var * tp
      |	ValDecl of var * term 
      | BoxDecl of var * term 
      | DiaDecl of var * term

    and lterm =
	LEmpty of location
      | LPrint of lterm * location
      |	LNat of int * location
      | LTrue of location
      | LFalse of location
      | LNil of location
      | LCons of lterm * lterm * location
      | LCar of lterm * location
      | LCdr of lterm * location
      | LVar of var * location
      | LBinOp of binop * lterm * lterm * location
      | LIf of lterm * lterm * lterm * location
      | LFn of var * lterm * location
      | LFix of var * var * lterm * location
      | LApp of lterm * lterm * location
      | LPair of lterm * lterm * location
      | LFst of lterm * location
      | LSnd of lterm * location
      | LBox of lterm * location
      | LDia of lclos * location
      | LLet of ldecl list * lterm * location
      | LChoose of lterm * location
      | LNu of var * lterm * location
      | LExpSubst of (var * lterm * location) list * lterm * location
      | LRaise of var * lterm * location
      | LHandle of (var * var * lterm * location) list * lterm * location
      | LShift of var * var * lterm * location
      | LReset of var * lterm * location
      | LOfTp of lterm * ltp * location

    and ldecl = 
	LNameDecl of var * ltp * location
      |	LValDecl of var * lterm * location
      | LBoxDecl of var * lterm * location
      | LDiaDecl of var * lterm * location

    (* types for possibility *)

    and clos =
        Clos of (var * term) list * term 
      | CLet of decl list * clos 

    and lclos =
        LClos of (var * lterm * location) list * lterm * location
      | LCLet of ldecl list * lclos * location


    type subst = term Var.VarMap.map


    (*****
           Functions for necessity -- implementations 
     *****)


    fun termToString (e:term):string =
	(case e of
	     Empty => "()"
	   | Print e' => 
		 let val s = termToString e'
		 in
		     "print "^s
		 end
	   | Nat n => Int.toString n
	   | True => "true"
	   | False => "false"
	   | Nil => "nil"
	   | Cons (e1, e2) => 
		 let val s1 = termToString e1
		     val s2 = termToString e2
		 in
		     s1^" :: "^s2
		 end
	   | Car e1 => 
		 let val s1 = termToString e1
		 in
		     "hd "^s1
		 end
	   | Cdr e1 => 
		 let val s1 = termToString e1
		 in
		     "tl "^s1
		 end
	   | Var (x) => Var.toString x
	   | BinOp (b, e1, e2) => 
		 let val s1 = termToString e1
		     val s2 = termToString e2
		 in
		     case b of
			 PLUS => s1 ^ " + " ^ s2
		       | MINUS => s1 ^ " - " ^ s2
		       | MULT => s1 ^ " * " ^ s2
		       | EQL => s1 ^ " = " ^ s2
		 end
	   | If (e1, e2, e3) => 
		 let val s1 = termToString e1
		     val s2 = termToString e2
		     val s3 = termToString e3
		 in
		     "if " ^ s1 ^ " then " ^ s2 ^ "\n" ^ 
                     "else " ^ s3
		 end
	   | Fn (x, e) => 
		 let val sx = Var.toString x
		     val s = termToString e
		 in
		     "fn " ^ sx ^ ". " ^ s
		 end

	   | Fix (f, x, e) => 
		 let val sf = Var.toString f
		     val sx = Var.toString x
		     val s = termToString e
		 in
		     "fun " ^ sf ^ "("^sx^") =\n" ^ s
		 end

	   | App (e1, e2) => 
		 let val s1 = termToString e1
		     val s2 = termToString e2
		 in
		     "("^s1^")("^s2^")"
		 end
	   | Pair (e1, e2) => 
		 let val s1 = termToString e1
		     val s2 = termToString e2
		 in
		     "("^s1^", "^s2^")"
		 end
	   | Fst (e) => 
		 let val s = termToString e
		 in
		     "fst " ^ s
		 end
	   | Snd (e) => 
		 let val s = termToString e
		 in
		     "snd " ^ s
		 end
	   | Box e => 
		 let val s = termToString e
		 in 
		     "box ("^ s ^ ")"
		 end

	   | Dia f => 
		 let val s = closToString f
		 in 
		     "dia ("^ s ^ ")"
		 end

	   | Let (decls, e) => 
		 let fun printDecls decls = 
		     (case decls of
			  nil => ""
			| (d::ds) => 
			      let val s1 = declToString d
				  val s2 = printDecls ds
			      in
				  s1^"\n"^s2
			      end)
		     val s1 = printDecls decls
		     val s2 = termToString e
		 in
		     "let " ^ s1 ^
                     "in\n" ^
                           s2 ^ "\n" ^
                     "end\n"
		 end
	   | Choose (e) => 
		 let val s = termToString e
		 in
		     "choose ("^s^")"
		 end
	   | Nu (x, e) => 
		 let val sx = Var.toString x
		     val s = termToString e
		 in
		     "nu "^sx^". "^s
		 end
	   | ExpSubst (assigns, e) => 
		 let val ss = expsubstToString assigns
		     val s = termToString e
		 in
		     "<"^ss^">"^"("^s^")"
		 end

	   | Raise (x, e) => 
		 let val sx = Var.toString x
		     val s = termToString e
		 in 
		     "raise "^sx^" "^s
		 end

	   | Handle (handlers, e) => 
		 let 
		     fun handlersPrint ([]) = ""
		       | handlersPrint ([(x, y, e)]) =
			 let val sx = Var.toString x
			     val sy = Var.toString y
			     val s = termToString e
			 in
			     sx ^ " " ^ sy ^ "->" ^ s
			 end
		       | handlersPrint ((x, y, e)::handlers') =
			 let val sx = Var.toString x
			     val sy = Var.toString y
			     val s = termToString e
			     val ss = handlersPrint handlers'
			 in
			     sx ^ " " ^ sy ^ "->" ^ s ^ ", " ^ ss
			 end
		     val ss = handlersPrint handlers
		     val s = termToString e
		 in
		     "("^s^") handle <"^ss^">"
		 end			       

	   | Shift (x, k, e) => 
		 let val sx = Var.toString x
		     val sk = Var.toString k
		     val s = termToString e
		 in 
		     "shift "^sx^" "^sk^". "^s
		 end
	     
	   | Reset (x, e) => 
		 let val sx = Var.toString x
		     val s = termToString e
		 in
		     "reset "^sx^" "^s
		 end

	   | OfTp (e, A) => 
		 let val s = termToString e
		     val ts = Types.toString A
		 in
		     "(" ^ s ^ " : " ^ ts ^ ")"
		 end)


    and expsubstToString (assigns) =
	(case assigns of 
	     nil => ""
	   | [(x, e)] =>
		 let val sx = Var.toString x
		     val s = termToString e
		 in
		     sx ^ "->" ^ s
		 end
	   | ((x, e)::decls) =>
		 let val sx = Var.toString x
		     val s = termToString e
		     val ss = expsubstToString decls
		 in
		     sx ^ "->" ^ s ^ ", " ^ ss
		 end)


    and declToString (decl:decl):string =
	(case decl of 
	     NameDecl (x, tp) =>
		 let val sx = Var.toString x
		     val st = Types.toString tp
		 in
		     "name "^sx^":"^st
		 end
	   | ValDecl (x, e) => 
		 let val sx = Var.toString x
		     val s = termToString e
		 in
		     "val "^sx^" = "^s
		 end
	   | BoxDecl (x, e) => 
		 let val sx = Var.toString x
		     val s = termToString e
		 in
		     "box "^sx^" = "^s
		 end
	   | DiaDecl (x, e) => 
		 let val sx = Var.toString x
		     val s = termToString e
		 in
		     "dia "^sx^" = "^s
		 end)




    and toTerm (lt:lterm) : term = 
	(case lt of
	     LEmpty _ => Empty
	   | LPrint (lt1, _) => 
		 Print (toTerm lt1)
	   | LNat (x, _) => Nat x
           | LTrue _ => True
           | LFalse _ => False
	   | LNil _ => Nil
	   | LCons (lt1, lt2, _) => 
		 Cons (toTerm lt1, toTerm lt2)
	   | LCar (lt1, _) => 
		 Car (toTerm lt1)
	   | LCdr (lt1, _) => 
		 Cdr (toTerm lt1)
           | LVar (x, _) => Var x
           | LBinOp (p, lt1, lt2, _) =>
		 let val t1 = toTerm lt1
		     val t2 = toTerm lt2
		 in 
		     BinOp (p, t1, t2)
		 end
           | LIf (lt1, lt2, lt3, _) => 
		 let val t1 = toTerm lt1
		     val t2 = toTerm lt2
		     val t3 = toTerm lt3
		 in
		     If (t1, t2, t3)
		 end
	     
	   | LFn (x, lt1, _) => 
		 let val t1 = toTerm lt1
		 in
		     Fn (x, t1)
		 end

	   | LFix (f, x, lt1, _) => 
		 let val t1 = toTerm lt1
		 in
		     Fix (f, x, t1)
		 end

           | LApp (lt1, lt2, _) => 
		 let val t1 = toTerm lt1
		     val t2 = toTerm lt2
		 in
		     App (t1, t2)
		 end

           | LPair (lt1, lt2, _) =>
		 let val t1 = toTerm lt1
		     val t2 = toTerm lt2
		 in
		     Pair (t1, t2)
		 end

           | LFst (lt1, _) => 
		 let val t1 = toTerm lt1
		 in
		     Fst (t1)
		 end
		 
           | LSnd (lt1, _) => 
		 let val t1 = toTerm lt1
		 in
		     Snd (t1)
		 end

           | LBox (lt1, _) => 
		 let val t1 = toTerm lt1
		 in
		     Box (t1)
		 end

	   | LDia (lf1, _) => 
		 let val f1 = toClos lf1
		 in
		     Dia (f1)
		 end
	     
           | LLet (lds, lt1, _) => 
		 let val ds = map toDecl lds
		     val t1 = toTerm lt1
		 in
		     Let (ds, t1)
		 end

          | LChoose (lt1, _) => 
		 let val t1 = toTerm lt1
		 in
		     Choose (t1)
		 end

          | LNu (x, lt1, _) => 
		 let val t1 = toTerm lt1
		 in
		     Nu (x, t1)
		 end

          | LExpSubst (lasgns, lt1, _) => 
		 let fun toAsgns (lasgns) = 
		     (case lasgns of
			  nil => nil
			| ((x, lt, _)::lasgns') => 
			      let val asgns = toAsgns lasgns'
				  val t = toTerm lt
			      in
				  (x, t)::asgns
			      end)
		     val asgns = toAsgns lasgns
		     val t1 = toTerm lt1
		 in
		     ExpSubst (asgns, t1)
		 end

	  | LRaise (x, lt1, _) => 
		 let val t1 = toTerm lt1
		 in
		     Raise (x, t1)
		 end

	  | LHandle (lhandlrs, lt1, _) => 
		 let fun toHandlrs (lhandlrs) = 
		     (case lhandlrs of
			  nil => nil
			| ((x, y, lt, _)::lhandlrs') => 
			      let val handlrs = toHandlrs lhandlrs'
				  val t = toTerm lt
			      in
				  (x, y, t)::handlrs
			      end)
		     val handlrs = toHandlrs lhandlrs
		     val t1 = toTerm lt1
		 in
		     Handle (handlrs, t1)
		 end   

	  | LShift (x, k, lt1, _) => 
		 let val t1 = toTerm lt1
		 in
		     Shift (x, k, t1)
		 end

	  | LReset (x, lt1, _) => 
		 let val t1 = toTerm lt1
		 in
		     Reset (x, t1)
		 end
	
          | LOfTp (lt1, ltp1, _) =>
		 let val t1 = toTerm lt1
		     val tp1 = Types.fromLtp ltp1
		 in
		     OfTp (t1, tp1)
		 end)


    and toDecl (ldecl:ldecl):decl =
	(case ldecl of
	     LNameDecl (x, ltp, _) =>
		 let val tp = Types.fromLtp ltp
		 in
		     NameDecl (x, tp)
		 end
	   | LValDecl (x, lt, _) => 
		 let val t = toTerm lt
		 in
		     ValDecl (x, t)
		 end
	     
	   | LBoxDecl (x, lt, _) => 
		 let val t = toTerm lt
		 in
		     BoxDecl (x, t)
		 end
	     
	   | LDiaDecl (x, lt, _) => 
		 let val t = toTerm lt
		 in
		     DiaDecl (x, t)
		 end)



    and termLoc (lt:lterm) = 
	(case lt of (
	     LEmpty loc 
	   | LPrint (_, loc)
	   | LNat (_, loc) 
           | LTrue loc 
           | LFalse loc
	   | LNil loc
	   | LCons (_, _, loc)
	   | LCar (_, loc)
	   | LCdr (_, loc)
           | LVar (_, loc)
           | LBinOp (_, _, _, loc)
           | LIf (_, _, _, loc) 
	   | LFn (_, _, loc)
	   | LFix (_, _, _, loc)
           | LApp (_, _, loc)
           | LPair (_, _, loc) 
           | LFst (_, loc)
           | LSnd (_, loc)
           | LBox (_, loc)
           | LDia (_, loc)
           | LLet (_, _, loc)
           | LChoose (_, loc)
           | LNu (_, _, loc)
           | LExpSubst (_, _, loc)
	   | LRaise (_, _, loc)
	   | LHandle (_, _, loc)
	   | LShift (_, _, _, loc)
	   | LReset (_, _, loc)
           | LOfTp (_, _, loc)) => loc)



    and declLoc (ldecl:ldecl) =
	(case ldecl of ( 
	     LNameDecl (_, _, loc)
	   | LValDecl (_, _, loc) 
	   | LBoxDecl (_, _, loc)
	   | LDiaDecl (_, _, loc)) => loc)


    and substName (s : subst, v : var) = 
	(case (Var.VarMap.find (s, v)) of
		 SOME (Var x) => x
	       | NONE => v)


    and substTerm  (s : subst, e : term) =
	(case e of
	     Empty => Empty
	   | Print e' => Print (substTerm (s, e'))
	   | Nat n => Nat n
	   | True => True
	   | False => False
	   | Nil => Nil
	   | Cons (e1, e2) => 
		 Cons (substTerm (s, e1), substTerm (s, e2))
	   | Car (e1) => 
		 Car (substTerm (s, e1))
	   | Cdr (e1) => 
		 Cdr (substTerm (s, e1))
	   | Var y => 
		 (case Var.VarMap.find (s, y) of
		      SOME (e') => e'
		    | NONE => Var y)
	   | BinOp (b, e1, e2) => 
		 BinOp (b, substTerm (s, e1), substTerm (s, e2))

	   | If (e1, e2, e3) => 
		 If (substTerm (s, e1), substTerm (s, e2), substTerm (s, e3))

	   | Fn (x, e') => 
		 let val x' = Var.derived_var x
		     val s' = Var.VarMap.insert (s, x, Var x')
		 in
		     Fn (x', substTerm (s', e'))
		 end

	   | Fix (f, x, e') => 
		 let val f' = Var.derived_var f
		     val x' = Var.derived_var x
		     val s' = Var.VarMap.insert (s, f, Var f')
		     val s'' = Var.VarMap.insert (s', x, Var x')
		 in
		     Fix (f', x', substTerm (s'', e'))
		 end

	   | App (e1, e2) => 
		 App (substTerm (s, e1), substTerm (s, e2))

	   | Pair (e1, e2) => 
		 Pair (substTerm (s, e1), substTerm (s, e2))

	   | Fst (e') => 
		 Fst (substTerm (s, e'))

	   | Snd (e') => 
		 Snd (substTerm (s, e'))

	   | Box (e') => 
		 Box (substTerm (s, e'))

	   | Dia (f') => 
		 Dia (substClos (s, f'))

	   | Let (decls, e) => 
		 let val (s', decls') = substDecls (s, decls)
		 in
		     Let (decls', substTerm (s', e))
		 end

	   | Choose (e') => Choose (substTerm (s, e'))

	   | Nu (x, e') => 
		 let val x' = Var.derived_var x
		     val s' = Var.VarMap.insert (s, x, Var x')
		 in
		     Nu (x', substTerm (s', e'))
		 end

	   | ExpSubst (asgmts, e') => 
		 let val asgmts'' = 
		     map (fn (v, e'') => (substName (s, v), 
					  substTerm (s, e''))) asgmts
		     val e'' = substTerm (s, e')
		 in
		     ExpSubst (asgmts'', e'')
		 end

	   | Raise (x, e') => 
		 Raise (substName (s, x), substTerm (s, e'))

	   | Handle (asgmts', e') => 
		 let val asgmts'' = 
		     map (fn (x, y, e'') => 
			  let val y' = Var.derived_var y
			      val s' = Var.VarMap.insert (s, y, Var y')
			  in
			      (substName (s, x), y', substTerm (s', e''))
			  end) asgmts'
		     val e'' = substTerm (s, e')
		 in
		     Handle (asgmts'', e'')
		 end
	     
	   | Shift (x, k, e') => 
		 let val k' = Var.derived_var k
		     val s' = Var.VarMap.insert (s, k, Var k')
		 in
		     Shift (substName (s, x), k', substTerm (s', e'))
		 end

	   | Reset (x, e') => 
		 Reset (substName (s, x), substTerm (s, e'))
	     
	   | OfTp (e', A) => 
		 OfTp (substTerm (s, e'), substType (s, A)))


    and substType (s : subst, tp:tp) : tp =
	(case tp of
	     UNIT => UNIT
	   | NAT => NAT
	   | BOOL => BOOL
	   | NATLIST => NATLIST
	   | PROD (tp1, tp2) => 
		 PROD (substType (s, tp1), substType (s, tp2))
	   | ARROW (tp1, tp2) => 
		 ARROW (substType (s, tp1), substType (s, tp2))
	   | NARROW (tp1, tp2) => 
		 NARROW (substType (s, tp1), substType (s, tp2))
	   | BOX (C, tp') => 
		 BOX (substSupport (s, C), substType (s, tp'))
	   | DIA (C, tp') => 
		 DIA (substSupport (s, C), substType (s, tp')))


    and substSupport (s : subst, C : support) : support =
	let val C' = 
	    map (fn x => substName (s, x)) 
	    (Support.toList C)
	in
	    Support.fromList C'
	end
	

    and substDecls (s : subst, decls : decl list) : subst * (decl list) =
	(case decls of
	     nil => (s, decls)
           | (NameDecl (x, tp)::ds) =>
		 let val tp' = substType (s, tp)
		     val x' = Var.derived_var x
		     val s' = Var.VarMap.insert (s, x, Var x')
		     val (s'', ds'') = substDecls (s', ds)
		 in
		     (s'', (NameDecl (x', tp')::ds''))
		 end
	   | (ValDecl (x, e)::ds) => 
		 let val e' = substTerm (s, e)
		     val x' = Var.derived_var x
		     val s' = Var.VarMap.insert (s, x, Var x')
		     val (s'', ds'') = substDecls (s', ds)
		 in
		     (s'', (ValDecl (x', e')::ds''))
		 end
	   | (BoxDecl (x, e)::ds) => 
		 let val e' = substTerm (s, e)
		     val x' = Var.derived_var x
		     val s' = Var.VarMap.insert (s, x, Var x')
		     val (s'', ds'') = substDecls (s', ds)
		 in
		     (s'', (BoxDecl (x', e')::ds''))
		 end
	   | (DiaDecl (x, e)::ds) => 
		 let val e' = substTerm (s, e)
		     val x' = Var.derived_var x
		     val s' = Var.VarMap.insert (s, x, Var x')
		     val (s'', ds'') = substDecls (s', ds)
		 in
		     (s'', (DiaDecl (x', e')::ds''))
		 end)


    and declEq (decl1, decl2) : subst * subst * bool =
	(case (decl1, decl2) of
	     (NameDecl (x, tp1), NameDecl (y, tp2)) => 
		 let val xy = Var.fromString_unique ("internal")
		     val s1 = Var.VarMap.insert (Var.VarMap.empty, x, Var xy)
		     val s2 = Var.VarMap.insert (Var.VarMap.empty, y, Var xy)
		 in
		     (s1, s2, Types.equal (tp1, tp2))
		 end
	   | ((ValDecl (x, e1'), ValDecl (y, e2')) | 
	      (BoxDecl (x, e1'), BoxDecl (y, e2')) | 
	      (DiaDecl (x, e1'), DiaDecl (y, e2'))) => 
		     let val xy = Var.fromString_unique ("internal")
			 val s1 = Var.VarMap.insert (Var.VarMap.empty, x, Var xy)
			 val s2 = Var.VarMap.insert (Var.VarMap.empty, y, Var xy)
		     in
			 (s1, s2, termEq (e1', e2'))
		     end)


    and termEq (e1, e2) =
	(case (e1, e2) of
	     (Empty, Empty) => true
	   | (Print e1, Print e2) => 
		 termEq (e1, e2)
	   | (Nat x, Nat y) => x = y
	   | (True, True) => true
	   | (False, False) => true
	   | (Nil, Nil) => true
	   | (Cons (hd1, tl1), Cons (hd2, tl2)) => 
		 termEq (hd1, hd2) andalso 
		 termEq (tl1, tl2)
	   | (Car e1, Car e2) => 
		 termEq (e1, e2)
	   | (Cdr e1, Cdr e2) => 
		 termEq (e1, e2)
	   | (Var x, Var y) => Var.equal (x, y)
	   | (BinOp (p1, e1, e2), BinOp (p2, e1', e2')) => 
		 ((p1 = p2) andalso 
		  termEq (e1, e1') andalso 
		  termEq (e2, e2'))
	   | (If (e1, e2, e3), If (e1', e2', e3')) => 
		 (termEq (e1, e1') andalso
		  termEq (e2, e2') andalso
		  termEq (e3, e3'))
	   | (Fn (x, e1), Fn (y, e2)) => 
		 let val xy = Var.fromString_unique ("internal")
		     val s1 = Var.VarMap.insert (Var.VarMap.empty, x, Var xy)
		     val s2 = Var.VarMap.insert (Var.VarMap.empty, y, Var xy)
		 in
		     termEq (substTerm (s1, e1), substTerm (s2, e2))
		 end
  	   | (Fix (f, x, e1), Fix (g, y, e2)) => 
		 let val xy = Var.fromString_unique ("internal1")
		     val fg = Var.fromString_unique ("internal2")
		     val s1 = Var.VarMap.insert (Var.VarMap.empty, f, Var fg)
		     val s1 = Var.VarMap.insert (s1, x, Var xy)
		     val s2 = Var.VarMap.insert (Var.VarMap.empty, g, Var fg)
		     val s2 = Var.VarMap.insert (s2, y, Var xy)
		 in
		     termEq (substTerm (s1, e1), substTerm (s2, e2))
		 end 

	   | (App (e1, e2), App(e1', e2')) => 
		 termEq (e1, e1') andalso 
		 termEq (e2, e2')
	   | (Pair (e1, e2), Pair (e1', e2')) => 
		 termEq (e1, e1') andalso termEq (e2, e2')
	   | (Fst e1, Fst e2) => 
		 termEq (e1, e2)
	   | (Snd e1, Snd e2) => 
		 termEq (e1, e2)
	   | (Box e1, Box e2) => 
		 termEq (e1, e2)
	   | (Dia f1, Dia f2) => 
		 closEq (f1, f2) 
	   | (Let (nil, e1), Let (nil, e2)) => 
		 termEq (e1, e2)
	   | (Let (decl1::ds1, e1), Let (decl2::ds2, e2)) => 
		 let val (s1, s2, b) = declEq (decl1, decl2)
		 in
		     b andalso 
		     termEq (substTerm (s1, Let (ds1, e1)),
			     substTerm (s2, Let (ds2, e2)))
		 end
	   | (Choose e1, Choose e2) => 
		 termEq (e1, e2)
	   | (Nu (x, e1), Nu (y, e2)) => 
		 let val xy = Var.fromString_unique ("internal")
		     val s1 = Var.VarMap.insert (Var.VarMap.empty, x, Var xy)
		     val s2 = Var.VarMap.insert (Var.VarMap.empty, y, Var xy)
		 in
		     termEq (substTerm (s1, e1), substTerm (s2, e2))
		 end

	   | (ExpSubst (asgmt1, e1), ExpSubst (asgmt2, e2)) => 
		 let fun asgmtEq (nil, nil) = true
		       | asgmtEq ((x,e1)::asg1, (y,e2)::asg2) =
		            Var.equal (x, y) andalso termEq (e1, e2)
			    andalso asgmtEq (asg1, asg2)
		 in
		     asgmtEq (asgmt1, asgmt2) andalso termEq (e1, e2)
		 end

	   | (Raise (x, e1), Raise (y, e2))  => 
		 Var.equal (x, y) andalso 
		 termEq (e1, e2)

	   | (Handle (asgmt1, e1), Handle (asgmt2, e2)) => 
		 let fun asgmtEq (nil, nil) = true
		       | asgmtEq ((x1,y1,e1)::asg1, (x2,y2,e2)::asg2) =
		            if Var.equal (x1, x2) then 
				let val y12 = Var.fromString_unique ("internal")
				    val s1 = Var.VarMap.insert (Var.VarMap.empty, y1, Var y12)
				    val s2 = Var.VarMap.insert (Var.VarMap.empty, y2, Var y12)
				in
				    termEq (substTerm (s1, e1), substTerm (s2, e2)) andalso
				    asgmtEq (asg1, asg2)
				end
			    else false
		 in
		     asgmtEq (asgmt1, asgmt2) andalso termEq (e1, e2)
		 end
	     
	   | (Shift (x, f, e1), Shift (y, g, e2)) => 
		 if Var.equal (x, y) then 
		     let val fg = Var.fromString_unique ("internal")
			 val s1 = Var.VarMap.insert (Var.VarMap.empty, f, Var fg)
			 val s2 = Var.VarMap.insert (Var.VarMap.empty, g, Var fg)
		     in
			 termEq (substTerm (s1, e1), substTerm (s2, e2))
		     end
		 else false
	   | (Reset (x, e1), Reset (y, e2)) => 
		 Var.equal (x, y) andalso termEq (e1, e2)
		 
	   | (OfTp (e1, tp1), OfTp (e2, tp2)) => 
		 termEq (e1, e2) andalso Types.equal (tp1, tp2)
	   | _ => false)
			 
		 


    (* Possibility functions -- implementation *)

    and closToString (f : clos) =
	(case f of
	     Clos (nil, e) => 
		 let val es = termToString e
		 in
		     "["^es^"]"
		 end
	   | Clos (assigns, e) => 
		 let val exps = expsubstToString assigns
		     val es = termToString e
		 in
		     "[<"^exps^">, "^es^"]"
		 end
	   | CLet (fdecls, f) => 
		 let fun printDecls fdecls = 
		     (case fdecls of
			  nil => ""
			| (d::ds) => 
			      let val s1 = declToString d
				  val s2 = printDecls ds
			      in
				  s1^"\n"^s2
			      end)
		     val s1 = printDecls fdecls
		     val s2 = closToString f
		 in
		     "let " ^ s1 ^
                     "in\n" ^
		     s2 ^ "\n" ^
                     "end\n"
		 end)	 

(*
    and cdeclToString (fdecl:decl):string =
	(case fdecl of 
	     CNameDecl (x, tp) =>
		 let val sx = Var.toString x
		     val st = Types.toString tp
		 in
		     "name "^sx^":"^st
		 end
	   | CValDecl (x, e) => 
		 let val sx = Var.toString x
		     val s = termToString e
		 in
		     "val "^sx^" = "^s
		 end
	   | CBoxDecl (x, e) => 
		 let val sx = Var.toString x
		     val s = termToString e
		 in
		     "box "^sx^" = "^s
		 end
	   | CDiaDecl (x, e) => 
		 let val sx = Var.toString x
		     val s = termToString e
		 in
		     "dia "^sx^" = "^s
		 end)
*)


    and toClos (f : lclos) : clos =
	(case f of
	     LClos (lexpsubst, lt, _) =>
		 let fun toAsgns (lasgns) = 
		     (case lasgns of
			  nil => nil
			| ((x, lt, _)::lasgns') => 
			      let val asgns = toAsgns lasgns'
				  val t = toTerm lt
			      in
				  (x, t)::asgns
			      end)
		     val expsubst = toAsgns lexpsubst
		     val t = toTerm lt
		 in
		     Clos (expsubst, t)
		 end
	   | LCLet (lfdecls, lf, _) => 
		 let val fdecls = map toDecl lfdecls
		     val f = toClos lf
		 in
		     CLet (fdecls, f)
		 end)


(*
    and toCDecl (lfdecl:ldecl):decl =
	(case lfdecl of
	     LCNameDecl (x, ltp, _) =>
		 let val tp = Types.fromLtp ltp
		 in
		     CNameDecl (x, tp)
		 end
	     
	   | LCValDecl (x, lt, _) => 
		 let val t = toTerm lt
		 in
		     CValDecl (x, t)
		 end
	     
	   | LCBoxDecl (x, lt, _) => 
		 let val t = toTerm lt
		 in
		     CBoxDecl (x, t)
		 end
	   | LCDiaDecl (x, lt, _) => 
		 let val t = toTerm lt
		 in
		     CDiaDecl (x, t)
		 end)
*)

    and closLoc (lf : lclos) : location =
	(case lf of (
	     LClos (_, _, loc) 
	   | LCLet (_, _, loc)) => loc)


(*    and cdeclLoc (lfdecl:ldecl) =
	(case lfdecl of (     
	     LNameDecl (_, _, loc)
	   | LValDecl (_, _, loc) 
	   | LBoxDecl (_, _, loc)
	   | LDiaDecl (_, _, loc)) => loc)
*)

    and substClos  (s : subst, f : clos) =
	(case f of
	     Clos (expsubst, e) => 
		 let val expsubst'' = 
		     map (fn (v, e'') => (substName (s, v), substTerm (s, e''))) expsubst
		     val e'' = substTerm (s, e)
		 in
		     Clos (expsubst'', e'')
		 end
	   | CLet (fdecls, f) => 
		 let val (s', fdecls') = substDecls (s, fdecls)
		 in
		     CLet (fdecls', substClos (s', f))
		 end)

(*

    and substCDecls (s : subst, fdecls : decl list) : subst * (decl list) =
	(case fdecls of
	     nil => (s, fdecls)  
	   | (NameDecl (x, tp)::ds) =>
		 let val tp' = substType (s, tp)
		     val x' = Var.derived_var x
		     val s' = Var.VarMap.insert (s, x, Var x')
		     val (s'', ds'') = substCDecls (s', ds)
		 in
		     (s'', (NameDecl (x', tp')::ds''))
		 end
	   | (ValDecl (x, e)::ds) => 
		 let val e' = substTerm (s, e)
		     val x' = Var.derived_var x
		     val s' = Var.VarMap.insert (s, x, Var x')
		     val (s'', ds'') = substCDecls (s', ds)
		 in
		     (s'', (ValDecl (x', e')::ds''))
		 end
	   | (BoxDecl (x, e)::ds) => 
		 let val e' = substTerm (s, e)
		     val x' = Var.derived_var x
		     val s' = Var.VarMap.insert (s, x, Var x')
		     val (s'', ds'') = substCDecls (s', ds)
		 in
		     (s'', (BoxDecl (x', e')::ds''))
		 end
	   | (DiaDecl (x, e)::ds) => 
		 let val e' = substTerm (s, e)
		     val x' = Var.derived_var x
		     val s' = Var.VarMap.insert (s, x, Var x')
		     val (s'', ds'') = substCDecls (s', ds)
		 in
		     (s'', (DiaDecl (x', e')::ds''))
		 end)
*)



    and closEq (f1, f2) =
	(case (f1, f2) of
	     (Clos (es1, e1), Clos(es2, e2)) => 
		 let fun asgmtEq (nil, nil) = true
		       | asgmtEq ((x,e1)::asg1, (y,e2)::asg2) =
		     Var.equal (x, y) andalso termEq (e1, e2)
		     andalso asgmtEq (asg1, asg2)
		 in
		     asgmtEq (es1, es2) andalso termEq (e1, e2)
		 end

	   | (CLet (nil, f1), CLet (nil, f2)) => 
		 closEq (f1, f2)

	   | (CLet (decl1::ds1, f1), CLet (decl2::ds2, f2)) => 
		 let val (s1, s2, b) = declEq (decl1, decl2)
		 in
		     b andalso
		     closEq (substClos (s1, CLet (ds1, f1)), 
			     substClos (s2, CLet (ds2, f2)))
		 end)
	     

end