(*Abstract type of variables*)
(*Leaf Petersen for CS15711 - Fall 97*)

signature VAR = 
  sig
    type var
      
    (* Maps variables to arbitrary data*)
    structure VarMap : ORD_MAP 
      where type Key.ord_key = var

    (* Sets of variables *)
    structure VarSet : ORD_SET
      where type Key.ord_key = var
      
    val fromString : string -> var
    (* Create a variable from a string.  The result is uniquely determined 
     * by the string - that is, equal(fromString s,fromString t) => true iff s=t.
     *)

    val fromString_unique : string -> var
    (* Create a variable from a string.  The result is guaranteed to be different
     * from all other variables - that is, for all s,t,
     * equal(fromString s,fromString t) => false.
     *)

    val new : unit -> var
    (* Generates an abritrary (non-unique) variable
     *)

    val derived_var : var -> var
    (* Given a variable v, 
     * equal(derived_var v,v'') => false for all v''.
     * derived_var = fromString_unique o toString
     * derived_var has the advantage that it maintains the same string
     * field, so that printed vars remain comprehensible.
     *)

    val compare : (var*var) -> order
    (* compare variables according to an arbitrary (not lexicographic!) 
     * ordering.
     *)

    val equal : (var*var) -> bool
    (* equal(v,v') returns true iff v and v' came from the same call to 
     * fromString_unique, new, or derived_var, or v and v' came from calls
     * to fromString with the same argument.
     *)

    val toString : var -> string
    (* Return a string represenation of the variable
     *)
      
  end

(* Implementation *)
structure Var :> VAR = 
  struct
    type var = {s:string,
		i:int}

    val counter = ref 0

    fun tick () = 
      let
	val x = !counter
      in
	(counter := x + 1;
	 x)
      end
		  
    fun fromString s = {s = s,i= (!counter)}
    fun fromString_unique s = {s = s,i= tick()}
    fun toString {s,i} = s (* ^"__"^(Int.toString i) *)
    fun new () = fromString "internal_var"
      
    fun derived_var {s,i} = {s=s,i=tick()}

    fun compare ({s=s1,i=i1},{s=s2,i=i2}) = 
      (case Int.compare (i1,i2)
	 of EQUAL => String.compare (s1,s2)
	  | other => other)

    fun equal ({s=s1,i=i1},{s=s2,i=i2}) = (i1=i2) andalso (s1=s2)

    structure VarKey : ORD_KEY = 
      struct
	type ord_key = var
	val compare = compare
      end

    structure VarSet : ORD_SET = SplaySetFn (VarKey)    
    structure VarMap : ORD_MAP = LocalSplayMapFn(VarKey)

  end
